# Defines the trigger class

import copy
from config import display_trigger
from _globals import *
from Triggers.Actions import *
from Triggers.Conditions import *
from Classes.class_tempswitch import TempSwitch

class T:
    def __init__(self, **inputs):
        # Players.  For example:
        # T( P=["Player 1", "Force 3"], C=[conditions], A=[actions] )
        # If you have a trigger inside another trigger's actions or conditions, not including player
        # definition will cause the imbedded trigger to simply inherit the containing trigger's player.
        # For example:
        # T(    P=[players], 
        #       C=[conditions], 
        #       A=[     actions,
        #               T( C=[other conditions], A=[other actions] ) ]
        #   )
        
        output = True
        if 'P' in inputs:
            self.P = inputs['P']
            for idx, player in enumerate(self.P):
                if isinstance(player, Player):
                    self.P[idx] = player.name
        else:
            self.P = None
            output = False
        
        
        # Conditions.  Separate the different conditions by commas.  Pretty straightforward.
        # The interesting thing you may do with TriGen Conditions is you can create custom conditions.
        # This means you may put triggers inside your conditions.
        # For example, lets say you want to see if unit1 and unit2 are within a location's distance of one another.
        # To create a condition which checks this you would do the following:
        #
        #   def close_together(unit1, unit2, location):
        #       return [T( A=[move location on unit1] ),
        #               T( C=[unit 2 is at the location], A=[True]  )]
        #
        #   T(   P=[player], 
        #        C=[close_together(unit1, unit2, some location), other conditions],
        #        A=[do actions]
        #    )
        # The 'True' keyword returns to TriGen that the custom condition you just created is satisfied.
        #
        # Also note that if you don't include any conditions, the condition always() is automatically inserted.
        #
        #
        #
        # Another cool feature of TriGen conditions is the ability to do 'and' and 'or' operations.
        # Use "[" and "]" for parenthesis and 'o' for or.
        # Here is an example:
        #
        # C=[[condition 1, condition 2, 'o', condition 3], [condition 4, 'o', condition 5]]
        #
        # Is logically equivalent to:
        #
        # C=[ ((condition 1 AND condition 2) OR condition 3) AND (condition 4 OR condition 5) ]
        #
        #
        # Finally there is a not keyword, 'n'.  It applies to the condition directly after the 'n'.
        
        self.C = [always()]
        if 'C' in inputs:
            self.C = inputs['C']
        
        # Actions.  One notable property of TriGen Actions:
        # You may place triggers inside of actions to perform condition checks within your overall trigger.
        # See above example for Players.
        
        self.A = []
        self.multiple_action_blocks = False
        if 'A' in inputs:
            if len(inputs['A']) > 62:
                self.multiple_action_block_switch = TempSwitch()
                self.A = inputs['A'][:62]
                self.A.append(self.multiple_action_block_switch.set())
                self.multiple_action_blocks = True
                self.rest_of_actions = inputs['A'][62:]
                #print(self.rest_of_actions)
            else:
                self.A = inputs['A']
        
        # Include the line just_once = True in the trigger to make it not have the preserve_trigger() action.
        # for example:
        # T( P=[player], C=[condition], A=[action], just_once = True )
        
        self.Just_Once = False
        if 'just_once' in inputs:
            self.Just_Once = inputs['just_once']
        
        # Else statements.  The structure is like the following:
        # T( P=[players], C=[conditions], A=[this executes if conditions are met], E=[this executes if conditions are not met] )
        #
        # To do if else statements simply put a trigger as inside the E.  For example:
        # T(    P=[players], 
        #       C=[conditions], 
        #       A=[this executes if conditions are met], 
        #       E=[     T(  C=[new conditions], 
        #                   A=[do this if new conditions are met])  ]
        #   )
        #
        # To do regular else statments just put actions in E.
        # T(    P=[players], 
        #       C=[conditions], 
        #       A=[this executes if conditions are met], 
        #       E=[  actions here go if conditions are not met  ]
        #   )
        
        self.E = None
        if 'E' in inputs:
            if isinstance(inputs['E'], list):
                self.E = T( A=inputs['E'] )
            else:
                self.E = inputs['E']
            
        # Special / Success. Used for after a custom condition is run.
        # The actions within Success are executed if every condition of that trigger is satisfied.
        # i.e. it places these actions into the action of the overall trigger.
        # This is useful if for example after running a custom condition you want to always execute a certain trigger.
        # For example, lets say deathcounter represents a button being pressed.
        # After a trigger detects this button being pressed you want to set the deathcounter to 0 so other triggers
        # which involve pressing that button don't run.
        # 
        # Here is an example:
        #   def button_is_pressed(num):
        #       global dc
        #       return T( C=[dc == num], A=[True], S=[dc.set_to(0)] )
        #
        #   T( P=[players], C=[button_is_pressed(7), other conditions], A=[actions] )
        
        self.S = None
        if 'S' in inputs:
            self.S = inputs['S']
            
        # Discretizer:
        
        if 'D' in inputs:
            if isinstance(inputs['D'], list):
                self.get_modified_by(inputs['D'][0])
            else:
                self.get_modified_by(inputs['D'])
                
        #internal
        self.in_middle_of_writing_conditions = False
        self.in_middle_of_writing_actions = False
        self.switches_to_destroy = []
        
        # output
        
        if output:
            return self.out()
            
        
            
    def get_modified_by(self, D):
        D.num = D.num + 1
        
        temp = copy.deepcopy(self.C)
        
        self.remove_specials(temp)
        
        #This sets D back to 0 so that the trigger can run again
        T( P=self.P, C=[D == D.num, 'n', temp], A=[D.set_to(0)] )
        
        self.C.insert(0, D == 0)
        
        self.A.append(D.set_to(D.num))
    
    def remove_specials(self, temp):
        all_specials = []
        for condition in temp:
            if isinstance(condition, list):
                self.remove_specials(condition)
            elif isinstance(condition, T):
                if condition.S != None:
                    all_specials.append(condition.S)
                    condition.S = None
        return
        
        
    # Creates the trigger
    def out(self):
        global file
        
        if self.E != None:
            else_switch = TempSwitch()
            self.A.append(else_switch.set())
            self.E.P = self.P
            if len(self.E.C) == 1:
                self.E.C = []
            self.E.C.insert(0, else_switch.is_clear())
        
        self.write_conditions(self.C)
        self.write_actions(self.A)
        
        
        
        if self.E != None:
            self.E.out()
            T( P=self.P, C=[else_switch.is_set()], A=[else_switch.clear()] )
            else_switch.destroy()
        
        for switch in self.switches_to_destroy:
            T( P=self.P, C=[switch.is_set()], A=[switch.clear()] )
            switch.destroy()
        
        return
    
    def end_trigger_with(self, action):
        self.start_actions()
        
        self.write_single_action(action)
        
        self.end_trigger()
        return
        
    def write_single_condition(self, condition):
        file.write("\t")
        file.write(condition)
        file.write("\n")
    
    def parse_list(self, list_to_check):
        parsed_list = [[]]
        k = 0
        
        for idx, condition in enumerate(list_to_check):
            if isinstance(condition, str):
                if condition == 'o':
                    k += 2
                    parsed_list.append('o')
                    parsed_list.append([])
                    
                elif condition == 'n':
                    if idx == 0:
                        #if isinstance(list_to_check[idx+1], T):
                        #    list_to_check[idx+1] = [[ T( C=list_to_check[idx+1].C, A=[list_to_check[idx+1].A, False], E=[True] ) ]]
                        #else:
                        list_to_check[idx+1] = [T( C=[list_to_check[idx+1]], A=[False], E=[True] )]
                    
                    else:
                        #if isinstance(list_to_check[idx+1], T):
                        #    list_to_check[idx+1] = [ T( C=list_to_check[idx+1].C, A=[list_to_check[idx+1].A, False], E=[True] ) ]
                        #else:
                        list_to_check[idx+1] = T( C=[list_to_check[idx+1]], A=[False], E=[True] )
                    
                else:
                    parsed_list[k].append(condition)
                    
            elif isinstance(condition, list):
                if k == 0 and ('o' not in condition):
                    for thing in condition:
                       parsed_list[k].append(thing)
                else:
                    parsed_list[k].append(condition)
            else:
                parsed_list[k].append(condition)
        
        if k == 0 and isinstance(parsed_list[0], list):
            parsed_list = parsed_list[0]
            if len(parsed_list) > 1:
                parsed_list = [[trigger] if isinstance(trigger, T) else trigger for trigger in parsed_list]
                
        
        return parsed_list
    
    def write_conditions(self, list_to_check, switch_to_output = None):
        #These switches will be outputted at the end.
        #Each will be required to be set in order for the actions to execute.
        switches_to_output = []
        
        #This is used for 'or' statements
        or_switch = None
        
        #This parses the list and puts brackets around everything between the ors.
        #For example: ["1", "2", 'o', "3"] then becomes [["1", "2"], 'o', ["3"]]
        #if isinstance(list_to_check[0], T) and len(list_to_check) == 1 and switch_to_output == None:
        #    list_to_check = [list_to_check]
        #else:
        list_to_check = self.parse_list(list_to_check)
        
        if isinstance(list_to_check[0], T) and len(list_to_check) == 1 and switch_to_output == None:
            list_to_check = [list_to_check]
            
        for condition in list_to_check:
            #If the condition is just a string, lets output it!
            
            if isinstance(condition, str):
                #Write the trigger if its not the 'o' handle.
                if condition != 'o':
                
                    #Don't write conditions without the trigger started first of course
                    if not self.in_middle_of_writing_conditions:
                        self.start_trigger()
                        
                    self.write_single_condition(condition)
                    
                #If the condition is the 'o' handle, then set the or_switch to the last switch used.
                #A switch is always used before the 'o' handle because a list always preceeds an 'o'
                #from the parsing process done earlier.
                else:
                    or_switch = switches_to_output[-1]
            
            #If the condition is a list, handle this separately.        
            elif isinstance(condition, list):
                
                
                
                #If this is the first list, i.e. you are in the middle of writing conditions,
                #then stop writing and output a "switch is set" action
                if self.in_middle_of_writing_conditions:
                    switches_to_output.append(TempSwitch())
                    self.end_trigger_with(switches_to_output[-1].set())
                    
                #If this is the first list you've come across, set up the or_switch.
                #(All other lists at this point will have or's before them)
                if or_switch == None:
                    
                    #There is an exception for if the condition contains only a trigger with no boolians in its actions
                    if len(condition) == 1 and isinstance(condition[0], T):
                        there_is_none = True
                        for action in condition[0].A:
                            if isinstance(action, bool):
                                there_is_none = False
                    else:
                        there_is_none = False
                    if there_is_none:
                        self.write_conditions(condition, "dummy")
                    else:
                        switches_to_output.append(TempSwitch())
                        self.write_conditions(condition, switches_to_output[-1])
                    
                #If this is not the first list, then it is a part of the 'or group'
                #i.e. if true, the trigger actions will execute.
                else:
                    if len(condition) == 1 and isinstance(condition[0], T):
                        there_is_none = True
                        for action in condition[0].A:
                            if isinstance(action, bool):
                                there_is_none = False
                    if there_is_none:
                        self.write_conditions(condition, "dummy")
                    else:
                        self.write_conditions(condition, or_switch)
            
            #Imbedded Triggers in conditions?!?!
            elif isinstance(condition, T):
                
                if condition.S != None:
                    self.A.append(condition.S)
                
                condition.P = self.P
                
                for idx, action in enumerate(condition.A):
                    if isinstance(action, bool):
                        there_is_none = False
                        if action:
                            condition.A[idx] = switch_to_output.set()
                        else:
                            condition.A[idx] = switch_to_output.clear()
                
                
                    
                
                if condition.E != None:
                    for idx, action in enumerate(condition.E.A):
                        if isinstance(action, bool):
                            if action:
                                condition.E.A[idx] = switch_to_output.set()
                            else:
                                condition.E.A[idx] = switch_to_output.clear()
                
                
                condition.out()
        
            
        self.switches_to_destroy += switches_to_output
        #If you have some switches that need to be set for the action to execute, output them to the file.
        if len(switches_to_output) > 0:
            if not self.in_middle_of_writing_conditions:
                self.start_trigger()
            for switch in switches_to_output:
                self.write_single_condition(switch.is_set())
                
        #This is used for lists.  If the conditions on the inside of a list are met,
        #this switch will be set.
        if switch_to_output != None and not isinstance(list_to_check[0], T):
            self.end_trigger_with(switch_to_output.set())
            
    
        return
        
                
    def start_trigger(self):
        global num_of_triggers
        num_of_triggers += 1
        file.write("Trigger(")
        
        # Players
        for idx, player in enumerate(self.P):
            was_list = False
            if type(player) is int:
                file.write("\"Player "+str(player)+"\"")
            elif type(player) is str:
                file.write("\""+player+"\"")
            elif type(player) is list:
                was_list = True
                for thing in player:
                    if isinstance(thing, Player):
                        self.P.append(thing.name)
                    else:
                        self.P.append(thing)
            if idx < len(self.P)-1 and not was_list:
                file.write(",")
        # Conditions
        file.write("){")
        file.write("\n")
        file.write("Conditions: ")
        file.write("\n")
        
        self.in_middle_of_writing_conditions = True
        for idx, player in enumerate(self.P):
            if isinstance(player, list):
                self.P.pop(idx)
        return
    
    def write_single_action(self, action):
        file.write("\t")
        file.write(action)
        file.write("\n")
        return
        
    def start_actions(self):
        self.in_middle_of_writing_conditions = False
        
        file.write("\n")
        file.write("Actions: ")
        file.write("\n")
        
        self.in_middle_of_writing_actions = True
        return
    
    def write_actions(self, list_to_check, end_trigger=True):
        switch_to_do_imbedded_trigger = None
    
        if not(self.in_middle_of_writing_actions):
            self.start_actions()
        
        for idx, action in enumerate(list_to_check):
            if isinstance(action, list):
                self.write_actions(action, False)
            if isinstance(action, str):
                self.write_single_action(action)
                
                
            # Imbedded Triggers in actions
            if isinstance(action, T):
                if switch_to_do_imbedded_trigger == None:
                    switch_to_do_imbedded_trigger = TempSwitch()
                    self.write_single_action(switch_to_do_imbedded_trigger.set())
                    self.end_trigger()
                
                action.P=self.P
                action.C = [switch_to_do_imbedded_trigger.is_set(), action.C]
                if action.E != None:
                    action.E.C = [switch_to_do_imbedded_trigger.is_set(), action.E.C]
                action.Just_Once = self.Just_Once
                action.out()
                
                
                #If the next index is not a trigger, Bool is False.  Otherwise Bool is true.
                bool = False
                if idx != len(list_to_check)-1:
                    if not isinstance(list_to_check[idx+1], T):
                        bool = True
                else:
                    bool = True
                    
                if bool:
                    self.start_trigger()
                    self.write_single_condition(switch_to_do_imbedded_trigger.is_set())
                    self.start_actions()
                    self.write_single_action(switch_to_do_imbedded_trigger.clear())
                    switch_to_do_imbedded_trigger.destroy()
                    switch_to_do_imbedded_trigger = None
                
        if end_trigger:
            self.end_trigger()
        return
        
    def end_trigger(self):
        if self.Just_Once == False:
            file.write("\t")
            file.write(preserve_trigger())
            file.write("\n")
        file.write("}")
        file.write("\n")
        file.write("\n")
        file.write("//-----------------------------------------------------------------//")
        file.write("\n")
        file.write("\n")
        
        self.in_middle_of_writing_actions = False
        
        if self.multiple_action_blocks:
            T(  P=self.P,
                C=[self.multiple_action_block_switch == "set"],
                A=self.rest_of_actions )
                
            
            T(  P=self.P,
                C=[self.multiple_action_block_switch == "set"],
                A=[self.multiple_action_block_switch.clear()] )
            self.multiple_action_block_switch.destroy()
        
        return
                
    
    @classmethod
    def num(clf):
        global num_of_triggers
        return num_of_triggers
        

        
        
        
        
        
        
        
        
        
        
        
        
